from dataclasses import dataclass
from typing import Any

from xcept import Exception_
from tests.tests.exception.helpers import pytest_message_parametrize


@pytest_message_parametrize
def test_behavior(template: str, test_field_value: str, expected_result: str) -> None:
    @dataclass
    class TestException(Exception_):
        test_field: Any

    exception = TestException(template, test_field=test_field_value)

    assert exception.get_message() == expected_result



from string import Formatter
from typing import Set

from xcept.formatter.helpers import get_field_from_expression


class HardFormatter(Formatter):

    def get_fields(self, string: str) -> Set[str]:
        fields = set()

        for i in self.parse(string):
            expression = i[1]

            if expression is not None:
                field, _ = get_field_from_expression(expression)
                fields.add(field)

        return fields



from typing import Optional, Tuple


def get_field_from_expression(expression: str) -> Tuple[str, Optional[str]]:
    for i in expression:
        if i in {".", "[", "!", ":"}:
            return expression.split(i, 1)[0], i

    field = _get_field_from_equal_sign_expression(expression)

    if field is not None:
        return field, "="

    return expression, None


def _get_field_from_equal_sign_expression(expression: str) -> Optional[str]:
    parts = expression.split("=")

    if len(parts) == 2:
        field, remainder = [i.strip(" ") for i in parts]

        if (
            not remainder
            and not field[0].isdigit()
            and all(
                i.isalpha()
                or i.isdigit()
                or i == "_"
                for i in field
            )
        ):
            return field



import dataclasses
from dataclasses import dataclass, field
from typing import Set
import warnings

from xcept.formatter.formatter import Formatter
from xcept.warnings import MissingFieldWarning, UnknownFieldWarning


@dataclass
class Exception_(Exception):  # noqa
    """Base class for custom non-exit exceptions.

    The behavior is mostly inherited from the built-in Exception class.
    The difference is that this class provides more convenience when
    working with custom exception attributes, and also allows you to
    create verbose messages using standard formatting syntax.

    If for some reason you don't need to include all attributes
    in a message, define ALL_REPLACEMENT_FIELDS_IS_REQUIRED = False
    to disable checks and warnings.
    """
    ALL_REPLACEMENT_FIELDS_IS_REQUIRED = True
    _template: str = field(repr=False, metadata={"is_template": True})

    def __post_init__(self):
        """Initialize an object.

        If a template does not contain all replacement fields and
        attribute ALL_REPLACEMENT_FIELDS_IS_REQUIRED is True, the
        xcept.warnings.MissingFieldWarning occurs.
        If a template contains unknown replacement fields, the
        xcept.warnings.UnknownFieldWarning occurs.
        """
        self._fields = {
            i.name
            for i in dataclasses.fields(self)
            if not i.metadata.get("is_template")
        }
        self._formatter = Formatter()
        replacement_fields = self._formatter.get_fields(self._template)

        if self.ALL_REPLACEMENT_FIELDS_IS_REQUIRED:
            _warn_about_missing_replacement_fields(
                fields={i for i in self._fields if i not in replacement_fields},
                template=self._template
            )

        _warn_about_unknown_replacement_fields(
            fields={i for i in replacement_fields if i not in self._fields},
            template=self._template
        )

    def __str__(self):
        """Return an exception message."""
        return self.get_message()

    def get_message(self) -> str:
        """Return an exception message."""
        arguments = {
            name: value
            for name, value in vars(self).items()
            if name in self._fields
        }

        return self._formatter.get_string(self._template, **arguments)


def _warn_about_missing_replacement_fields(fields: Set[str], template: str) -> None:
    if fields:
        warnings.warn(
            message=MissingFieldWarning(
                f"No the replacement {_get_field_message_part(fields)} "
                f"in the template {template!r}!"
            ),
            stacklevel=4
        )


def _warn_about_unknown_replacement_fields(fields: Set[str], template: str) -> None:
    if fields:
        warnings.warn(
            message=UnknownFieldWarning(
                f"Unknown the replacement {_get_field_message_part(fields)} "
                f"in the template {template!r}!"
            ),
            stacklevel=4
        )


def _get_field_message_part(fields: Set[str]) -> str:
    if len(fields) == 1:
        message_part = f"field {''.join(fields)!r}"
    elif len(fields) > 1:
        message_part = f"fields {', '.join(repr(i) for i in fields)}"
    else:
        raise ValueError("No fields!")

    return message_part



from typing import Set

from xcept.formatter.formatters.hard import HardFormatter
from xcept.formatter.formatters.soft import SoftFormatter


class Formatter:

    def __init__(self):
        self._hard_formatter = HardFormatter()

    # noinspection PyMethodMayBeStatic
    def get_string(self, template: str, **kwargs) -> str:
        return SoftFormatter(set(kwargs)).format(template, **kwargs)

    def get_fields(self, template: str) -> Set[str]:
        return self._hard_formatter.get_fields(template)



from string import Formatter
from typing import Optional, Set, List, Tuple

from xcept.formatter.helpers import get_field_from_expression


UNKNOWN_FIELD_VALUE = "<UNKNOWN>"


class SoftFormatter(Formatter):

    def __init__(self, fields: Set[str]):
        self._fields = fields

    def parse(
        self,
        format_string: str
    ) -> List[Tuple[str, Optional[str], Optional[str], Optional[str]]]:
        parts = []
        texts = []

        for text, field_expression, format_spec, conversion in super().parse(format_string):
            if field_expression is not None:
                field, separator = get_field_from_expression(field_expression)

                if separator == "=":
                    text += field_expression
                    field_expression = field
                    conversion = "r"

                if field in self._fields:
                    parts.append(("".join(texts) + text, field_expression, format_spec, conversion))
                    texts.clear()
                else:
                    texts.append(text + UNKNOWN_FIELD_VALUE)
            else:
                texts.append(text)

        if texts:
            parts.append(("".join(texts), None, None, None))

        return parts



import setuptools
from pathlib import Path


def get_long_description() -> str:
    path = Path(__file__).parent / "README.md"

    with path.open(encoding="UTF-8") as stream:
        long_description = stream.read()

    return long_description


setuptools.setup(
    name="xcept",
    version="3.0.0",
    packages=setuptools.find_packages(exclude=("tests",)),
    url="https://github.com/Abstract-X/xcept",
    author="Abstract-X",
    author_email="abstract-x-mail@protonmail.com",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    python_requires='>=3.7',
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11"
    ]
)



from dataclasses import dataclass

import pytest

from xcept import Exception_
from xcept.warnings import MissingFieldWarning, UnknownFieldWarning


@pytest.mark.parametrize(
    ("template",),
    (
        ("",),
        ("test_field",),
        ("{{test_field}}",),
        ("message",)
    )
)
def test_template_with_missing_replacement_fields(template: str) -> None:
    @dataclass
    class TestException(Exception_):
        test_field: str

    with pytest.warns(MissingFieldWarning):
        TestException(template, test_field="test_value")


@pytest.mark.parametrize(
    ("template",),
    (
        # Positional field without index
        ("{}",),
        ("{!r}",),
        ("{:.2f}",),

        # Positional field with index
        ("{0}",),
        ("{0!r}",),
        ("{0:.2f}",),
        ("{0[0]}",),
        ("{0.attribute}",),

        # Keyword field
        ("{test_field} {unknown_field}",),
        ("{test_field} {unknown_field!r}",),
        ("{test_field} {unknown_field:.2f}",),
        ("{test_field} {unknown_field[0]}",),
        ("{test_field} {unknown_field.attribute}",),
    )
)
def test_template_with_unknown_replacement_fields(template: str) -> None:
    @dataclass
    class TestException(Exception_):
        test_field: str

    with pytest.warns(UnknownFieldWarning):
        TestException(template, test_field="test_value")
